
package org.dronus.graph.gui;


import java.awt.Point;
import  java.util.*;
import java.io.*;
import org.dronus.graph.Graph;

/* Diese Klasse hlt eine "Locality", einfach 
 einen (gerade sichtbaren) Untergraphen des gesamten Graphen.
 
 Das wichtigste hier ist die Fhrung einer Gitter-Hash-Map, die fr die 
 Utility-Funktion "forNearNodePairs" siehe unten benutzt wird. Diese Funktion ist
 der eigentliche Nutzwert dieser Klasse.
   
*/
public class Locality extends Graph {

    protected Graph completeEltSet; //Verweis auf den ganzen Graphen

/*	Der heilige Gral: nodeMap ist eine gitterartige Hash-Tabelle aller Knoten dieser Locality
 * Die Gitterweite entspricht der zu nutzenden Maximalerntfernung fr
 * "forNearNodePairs" s.u. 
 * nodesDirty wird von Settern gesetzt, um die Notwendigkeit eines Updates vor dem
 * nchsten Benutzen zu markieren.
 * Die HashMap wird mit einer Load von 0.5f betrieben, da sie nicht sehr
 * gro ist, und hier Performance vor Speicher zhlt.    
 * */
	protected HashMap nodeMap =  new HashMap(100,0.5f);  
	boolean nodesDirty=true; //checks validity of nodeMap
  
  // ............der bliche kleinkram, einzige besonderheit die nodesDirty-Markierung 

  	public Locality(Graph ges) {
        super();
        completeEltSet = ges;
    }

	synchronized public Graph getCompleteEltSet() {
        return completeEltSet;
    }

	synchronized public void addNode( LayoutableNode n ) {
        if (!contains(n)) {
            super.addNode(n);
            nodesDirty=true;
            if (!completeEltSet.contains(n)) completeEltSet.addNode(n);
        }
    }

	synchronized public void addEdge( LayoutableEdge e ) {
        if(!contains(e)) {
            edges.add(e);
            if (!completeEltSet.contains(e)) completeEltSet.addEdge(e);
        }
    }

	synchronized public void addNodeWithEdges( LayoutableNode n ){
        addNode(n);
        for (int i = 0 ; i < n.edgeCount(); i++) {
            LayoutableEdge e=n.edgeAt(i);
            if(contains(e.getOtherEndpt(n))) addEdge(e);
        }

    }
    
  /*  public synchronized void addAll() throws TGException {
        synchronized (completeEltSet) {
        	nodes.addAll(completeEltSet.nodes);
			edges.addAll(completeEltSet.edges);            
        }
    }
    */	
	synchronized public LayoutableEdge findEdge( LayoutableNode from, LayoutableNode to ) {
        LayoutableEdge foundEdge=super.findEdge(from,to);
        if (foundEdge!=null && edges.contains(foundEdge)) return foundEdge;
        else return null;
    }

	synchronized public boolean deleteEdge( LayoutableEdge e ) {
        if (e == null) return false;
        else {
            removeEdge(e);
            return completeEltSet.deleteEdge(e);
        }
    }
	synchronized public boolean removeEdge( LayoutableEdge e ) {
        if (e == null) return false;
        return edges.remove(e);
    }
	synchronized public boolean deleteNode( LayoutableNode node ) {
        if ( node == null ) return false;
        else {
            removeNode(node);
            return completeEltSet.deleteNode(node);
        }
    }

	synchronized public boolean removeNode( LayoutableNode node ) {
          if (node == null) return false;
          if (!nodes.remove(node)) return false;
		  
		  String id = node.getID();
          if ( id != null ) nodeIDRegistry.remove(id); // remove from registry
		
          for (int i = 0 ; i < node.edgeCount(); i++) {
             removeEdge(node.edgeAt(i));
         }
			
		nodesDirty=true;
		
         return true;
    }

    public synchronized void removeNodes( List nodesToRemove ) {
        for (int i=0;i<nodesToRemove.size();i++) {
            removeNode((LayoutableNode) nodesToRemove.get(i));
        }
    }

    public synchronized void removeAll() {
        super.clearAll();
		nodesDirty=true;       
    }

    public synchronized void clearAll() {
        removeAll();	
        completeEltSet.clearAll();
    }


	//nun zum Kern
	
	/* builNodeMap wird benutzt, um einen Haufen Knoten in die Map zu ordnen. 
	 * Sollte auch statt "update" werden, wenn sich die Knotenmenge ndert 
	 * (dafr sorgt hier das nodeDirty)
	 * 
	 * RANGE gibt die Gitterweite an.
	 */
	void buildNodeMap(final double RANGE){		
		nodeMap=new HashMap(100,.5f);
		LayoutableNode n;
		Iterator i = nodes.iterator();
		/* ganz einfach: zu jedem Node wird ein Point-Objekt seiner auf RANGE 
		runterdividierten Integerkoordinaten konstruiert, welches eine HashFunktion	
		fr das Koordinatenpaar bereit stellt.
		Mgliche flash-lsung: String-Hash.
		*/
		while(i.hasNext() ){
			n=(LayoutableNode)i.next();
			//p: der Hash-fhige 2d-Integer-Punkt
			Point p = new Point((int)(n.x/RANGE),(int)(n.y/RANGE));
			//l: die Liste aller Nodes dieser Gitter-Zelle
			List  l=(List)nodeMap.get(p); 
			//falls noch gar kein Knoten in dieser Zelle: (dann fehlt auch der Vector)
			if (l==null) nodeMap.put(p,l=new ArrayList());
			//Und ab mit dem Knoten in die Zelle v. 
			l.add(n);			
		}
		//alles klar, Knotenmap ist aktuell.
		nodesDirty=false;		
	}
	
	/* updateNodeMap aktualisiert die Nodemap, solange die Knoten nur bewegt werden
	 * es kann keine Knotenmengennderungen vertragen.
	 * 
	 * Es wird einfach nachgesehen, ob jeder knoten auch in der Gitterzelle ist, 
	 * wo er seinen Koordinaten nach reingehrt. Wenn nicht, wird er verschoben.
	 */
	void updateNodeMap(double RANGE){
		//moveNodes: speichert die zu verschiebenden Nodes zwischen.
		List moveNodes=new ArrayList();
		Iterator k=nodeMap.keySet().iterator(); 
		//fr jede Gitterzelle 
		while (k.hasNext()){
			Point p1 = (Point)k.next();
			List v1=(List)nodeMap.get(p1);
			Iterator i=v1.iterator();
			//fr jeden Knoten in dieser Gitterzelle
			while ( i.hasNext() ){
				LayoutableNode n=(LayoutableNode)i.next();								
				//ist der Knoten auch in der Zelle, wo seine Koordinaten in hingehren lassen?
				if (p1.equals(new Point((int)(n.x/RANGE),(int)(n.y/RANGE)))) continue;				
				//nein: dann in den Zwischenspeicher, und aus der Zelle lschen.
				moveNodes.add(n);
				i.remove();							
			}
			//ist die Gitterzelle ganz leer geworden, erntferne sie.
			if (v1.isEmpty()) k.remove();  
		}
		//2. Teil: nun die verschobenen Knoten neu einfgen, analog zu "buildNodeMap" s.o.	
		Iterator m=moveNodes.iterator();
		while (m.hasNext()){
			LayoutableNode n=(LayoutableNode)m.next();
			Point p = new Point((int)(n.x/RANGE),(int)(n.y/RANGE));
			List l=(List)nodeMap.get(p); 
			if (l==null) nodeMap.put(p,l=new ArrayList());
			l.add(n);
		}
	}
	
	/** iterates through near pairs of Nodes. */
	/* nun die Utility-Funktion, fr die all die Vorbereitungen oben da sind:
	 * "TGForEachNodePair" ist eine Klasse, die der Nutzniesser mit einer Funktion 
	 * seiner Wahl ausstattet, die er auf Knotenpaaren verarbeiten muss.
	 * (z.b. die Abstossung der knoten)
	 * 
	 * "forNearNodePairs" fhrt die Funktion darin dann fr alle Knotenpaare mit einem 
	 * Knotenabstand unterhalb der euklidischen Distanz RANGE aus.
	 * 
	 */
	synchronized public void forNearNodePairs( final GraphForEachNodePair fenp, final double RANGE ) {	   
		int rangeChecks=0,executes=0; //Statistik
		double RANGE2=RANGE*RANGE;
		double MAPRANGE=RANGE/2;
		/* einen Monitor auf die Knoten einrasten, damit uns sie keiner
		 * vor unseren  Augen manipuliert.	*/
		
	
		if (nodesDirty) //wenn knoten gendert,  
			buildNodeMap(MAPRANGE); //nodeMap neu aufbauen
		else updateNodeMap(MAPRANGE);//sonst nur bewegungs-update s.o.
		Iterator it=nodeMap.keySet().iterator();
		
		//fr alle Gitterzellen: 
		while (it.hasNext()){
			final Point p = (Point)it.next();
			final List v1=(List)nodeMap.get(p);
			//ein Cross-Check aller Knoten innerhalb dieser Zelle 
			//(Symmetrie der Paare wird beachtet!)
			//Fr alle Knoten dieser Zelle
			final int v1s=v1.size();
			for (int i1=0; i1<v1s; i1++){
				final LayoutableNode  n1=(LayoutableNode)v1.get(i1);
				//Fr die ersten i1 Knoten derselben Zelle (jedes paar nur einmal)
				for (int i2=0; i2<i1; i2++){
					final LayoutableNode   n2=(LayoutableNode)v1.get(i2);
					final double vx=n1.x-n2.x, vy=n1.y-n2.y;
					rangeChecks++;
					//einige Shortcut-Tests, dann die euklidische Distanz.
					if (-RANGE<vx &&  vx<RANGE && -RANGE<vy && vy<RANGE && vx*vx+vy*vy<RANGE*RANGE){
						fenp.forEachNodePair(n1,n2); // Nah genug? Dann die bergebene Funktion ausfhren. 
						executes++;
					}
				 }
			 }			
			 		
			 List v2;
			 /* nun nochmal das ganze fr die Nachbarsektoren. 
			  * Die 9er-Nachbar reduziert sich zu vier Zellen,
			  * Die 25er zu 12, da die anderen durch berlappung erreicht
			  * werden. Sozusagen eine rechts-unten-nachbarschaft.
			  */
			 //definition der Nachbarn				 
			 //Point[] neighbours={new Point(p.x, p.y+1),	new Point(p.x+1, p.y+1),new Point(p.x+1, p.y), new Point(p.x+1,p.y-1)};
			final Point[] neighbours={	new Point(p.x, p.y+1), new Point(p.x+1, p.y+1),new Point(p.x+1, p.y), new Point(p.x+1,p.y-1),
										new Point(p.x-1, p.y+2), new Point(p.x, p.y+2),	new Point(p.x+1, p.y+2),new Point(p.x+2, p.y+2), new Point(p.x+2,p.y+1), new Point(p.x+2, p.y), new Point(p.x+2, p.y-1),new Point(p.x+2, p.y-2)};
			 //fr alle  rechts-unten-nachbarn
			 for (int in=0; in<neighbours.length; in++) 
			 	 //ist in dieser Nachbarzelle etwas?
				 if ((v2=(List)nodeMap.get(neighbours[in]))!=null)
				 	 //fr alle Knoten dieser Zelle
					 for (int i1=0; i1<v1.size(); i1++){ //so we cross-check the two sectors
						final LayoutableNode  n1=(LayoutableNode)v1.get(i1);
						/*	fr alle Knoten der Nachbarzelle
						* hier ist  keine Symmetriebeachtung wie oben ntig, 
						* da bereits die Wahl der Zellen-Nachbarn das bercksichtigt! 
						*/		
						final int v2s=v2.size();			
						for (int i2=0; i2<v2s; i2++){
							final LayoutableNode  n2=(LayoutableNode)v2.get(i2);
							 final double vx=n1.x-n2.x, vy=n1.y-n2.y;
							 rangeChecks++;
							 //distanztest s.o.
							 if (-RANGE<vx &&  vx<RANGE && -RANGE<vy && vy<RANGE && vx*vx+vy*vy<RANGE2){
								fenp.forEachNodePair(n1,n2);// Nah genug? Dann die bergebene Funktion ausfhren.
								executes++;
							 }
								 
						 }
					}
		}
			//bei bedarf: statistik ausgeben
			//int nc=nodeCount(); if (nc>0) System.out.println("MapCells: "+nodeMap.size()+" CellSize: "+nc/nodeMap.size()+" needed "+rangeChecks+" rangechecks, made "+executes+ " executes."); 
   }
   
   public void loadLocality(File f){
		   //URL u=new URL("file://test.mgl");
		   removeAll();
		   try {
			   BufferedReader s = new BufferedReader(new FileReader(f));
			   String l;
			   while((l=s.readLine())!=null){
				   StringTokenizer tok = new StringTokenizer(l);
				   String id = tok.nextToken();
				   LayoutableNode n=completeEltSet.findNode(id);
				   if (n==null) continue;
				   addNodeWithEdges(n);
				   n.x=(new Double(tok.nextToken())).doubleValue();
				   n.y=(new Double(tok.nextToken())).doubleValue();
				   n.fixed = ((new Boolean(tok.nextToken())).booleanValue());
				   n.stop();	
			   }			
		   } catch (Exception e) {
			   if (e instanceof FileNotFoundException) System.out.println("file not found:"+f.toString());  
			   else System.out.println("not a valid layout-file.");
			   e.printStackTrace();
		   }						
	   }

	   public void saveLocality(File f){
		   try {
			   final PrintWriter s = new PrintWriter(new FileWriter(f));
			   GraphForEachNode saver=new GraphForEachNode(){
					   public void forEachNode(LayoutableNode n) {
						   s.println(n.getID()+" "+n.x+" "+n.y+" "+n.fixed);
						   System.out.println(n.getID()+" "+n.x+" "+n.y+" "+n.fixed);
					   } 
			   };
			   forAllNodes(saver);
			   s.close();			
		   } catch (IOException e) {
			   System.out.print(e);
		   }	
	   }
}

